#include "tux.h"
#include "y.tab.h"
#include <stdio.h>

#define NSTACK 256

Datum stack[NSTACK];
Datum *stackp;

#define NPROG 2000
Inst	prog[NPROG];	/* the machine				*/
Inst	*progp;		/* next free spot for code gen 		*/
Inst	*pc;		/* program counter during exec 		*/
Inst	*progbase = prog;	/* start of current subprog 	*/
int	returning;	/* 1 if return stmt seen 		*/
typedef	struct Frame {	/* proc/func call stack frame 		*/
	Symbol	*sp;	/* symbol table entry			*/
	Inst	*retpc;	/* where to resume after return 	*/
	Datum	*argn;	/* n-th argment on stack 		*/
	int	nargs;	/* number of arguments 			*/
} Frame;

#define NFRAME 100
Frame 	frame[NFRAME];
Frame	*fp;		/* frame pointer			*/

initcode()		/* initialize for code generation 	*/
{
	progp  = progbase;
	stackp = stack;
	fp = frame;
	returning = 0;
}
void   displaystack(sp)
        Datum *sp;
{
        printf("+--------------data-stack------------+\n");
        for ( ; sp >= stack; *sp--) {
                printf(" > sp addr [%x] - >> [%x]\n", sp, *sp);
        }
        printf("+------------------------------------+\n");
}

push(d)			/* push d oto stack */
	Datum d;
{
	if (stackp >= &stack[NSTACK])
		execerror("stack overflow", (char *) 0);
	*stackp = d;
	*stackp++;
}
Datum pop()		/* pop/return top element from stack */
{
	if (stackp <= stack)
		execerror("stack underflow", (char *) 0);
	*stackp--;
	//printf("[[%x]]\n", *stackp);
	return *stackp;
}
Inst *code(f)		/* install one instruction or operand */
	Inst f;
{
	Inst *oprogp = progp;
	if (progp >= &prog[NPROG])
		execerror("program too big", (char *) 0);
	*progp++ = f;
	return oprogp;
}
execute(p)		/* run the machine */
	Inst *p;
{
	for (pc = p; *pc != STOP && !returning; ) {
#if DEBUG==1
		displaysym();
		displayprog();
		displaystack(stackp);
		printf("running code (exec) code[%x]\n", *pc);
#endif
		//Datum a = (*(*pc++))(); 		/* changed so works on both 32/64 bit machines */
		(*(*pc++))(); 		/* changed so works on both 32/64 bit machines */
#if DEBUG==1
		//printf("exiting code (exec)\n");
#endif
	}
}
constpush()		/* push constant onto stack */
{
	Datum d;
	d.val = ((Symbol *)*pc++)->u.val;
	push(d);
}
varpush()		/* push variable onto stack */
{
	Datum d;
	d.sym = (Symbol *)(*pc++);
	push(d);
}
int add()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val += d2.val;
	push(d1);
}
int sub()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val -= d2.val;
	push(d1);
}
int mul()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val *= d2.val;
	push(d1);
}
int divv()
{
	Datum d1, d2;
	d2 = pop();
	if (d2.val == 0.0)
		execerror(" division by zero ", (char *) 0);
	d1 = pop();
	d1.val /= d2.val;
	push(d1);
}
int negate()
{
	Datum d;
	d = pop();
	d.val = -d.val;
	push(d);
}
int power()
{
	Datum d1, d2;
	extern double Pow();
	d2 = pop();
	d1 = pop();
	d1.val = Pow(d1.val, d2.val);
	push(d1);
}
int eval()			/* evaluate variable on stack */
{
	Datum d;
	//printf("in eval\n");
	d = pop();
	if (d.sym->type == UNDEF)
		execerror("undefined variable", d.sym->name);
	d.val = d.sym->u.val;
	push(d);
	//printf("exit eval\n");
}
int assign()		/* assign top value to next value */
{
	Datum d1, d2;
	d1 = pop();
	d2 = pop();
	if (d1.sym->type != VAR && d1.sym->type != UNDEF)
		execerror("assignment to non-variable", d1.sym->name);
	d1.sym->u.val  = d2.val;
	d1.sym->type  = VAR;
	push(d2);
}
int print()
{
	Datum d;
	d = pop();
	printf("\t%.8g\n", d.val);
}
int bltin()
{
	Datum d;
	d = pop();
	d.val = (*(double (*)())(*pc++))(d.val);
	push(d);
}	
int gt()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val = (double)(d1.val > d2.val);
	push(d1);
}
int lt()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val = (double)(d1.val < d2.val);
	push(d1);
}
int ge()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val = (double)(d1.val >= d2.val);
	push(d1);
}
int le()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val = (double)(d1.val <= d2.val);
	push(d1);
}
int eq()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val = (double)(d1.val == d2.val);
	push(d1);
}
int ne()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val = (double)(d1.val != d2.val);
	push(d1);
}
int and()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val = (double)(d1.val != 0.0 &&  d2.val != 0.0);
	push(d1);
}
int or()
{
	Datum d1, d2;
	d2 = pop();
	d1 = pop();
	d1.val = (double)(d1.val != 0.0 ||  d2.val != 0.0);
	push(d1);
}
int not()
{
	Datum d;
	d = pop();
	d.val = (double)(d.val == 0.0);
	push(d);
}
int whilecode() 
{
	Datum d;
	Inst *savepc = pc;			/* body of loop */
	execute(savepc+2);			/* condition    */
	d = pop();
	while (d.val) {
		execute(*((Inst **)(savepc)));	/* body */
		if (returning)
			break;
		execute(savepc+2);
		d = pop();
	}
	if (!returning)
		pc = *((Inst **)(savepc+1));	/* next statement */
}
int ifcode()
{
	Datum d;
	Inst *savepc = pc;			/* then part */
	execute(savepc+3);			/* condition */
	d = pop();
	if (d.val)
		execute(*((Inst **)(savepc)));
	else if (*((Inst **)(savepc+1))) 	/* else part? */
		execute(*((Inst **)(savepc+1)));
	if (!returning)
		pc = *((Inst **)(savepc+2));	/* next stmt  */
}
int define(sp)		/* put func/proc in symbol table */
	Symbol *sp;
{
	sp->u.defn = (Inst)progbase;	/* start of code */
	progbase = progp;		/* next code starts here */
}
int call()		/* call a function */
{
	Symbol *sp = (Symbol *)pc[0];	/* symbol table entry */
					/* for function	      */
	if (fp++ >= &frame[NFRAME-1])
		execerror(sp->name, "call nested too deeply");
	fp->sp = sp;
	fp->nargs = (int)pc[1];
	fp->retpc = pc + 2;
	fp->argn = stackp - 1;		/* last argument */
	execute(sp->u.defn);
	returning = 0;
}
int ret()		/* common return from func or proc */
{
	int i;
	for (i = 0; i < fp->nargs; i++)
		pop();	/* pop args */
	pc = (Inst *)fp->retpc;
	--fp;
	returning = 1;
}
funcret()		/* return from a function */
{
	Datum d;
	if (fp->sp->type == PROCEDURE)
		execerror(fp->sp->name, "(proc) returns value");
	d = pop();	/* preserve function return value */
	ret();
	push(d);
}
procret()		/* return from a procedure */
{
	if (fp->sp->type == FUNCTION)
		execerror(fp->sp->name, "(func) returns no value");
	ret();
}
double *getarg()	/* return pointer to argument */
{
	int nargs = (int) *pc++;
	if (nargs > fp->nargs)
		execerror(fp->sp->name, "not enough arguments");
	return &fp->argn[nargs - fp->nargs].val;
}
int arg()		/* push arguments onto stack */
{
	Datum d;
	d.val = *getarg();
	push(d);
}
int argassign()		/* store top of stack in argument */
{
	Datum d;
	d = pop();
	push(d);	/* leave value on stack */
	*getarg() = d.val;
}
int prstr()		/* print string value  */
{
	printf("%s", (char *) *pc++);
}
int prexpr()		/* print numeric value */
{
	Datum d;
	d = pop();
	printf("%.8g\n", d.val);
}
int varread()		/* read into variable */
{
	Datum d;
	extern FILE *fin;
	Symbol *var = (Symbol *) *pc++;
 Again:
	switch (fscanf(fin, "%lf", &var->u.val)) {
	case EOF:
		if (moreinput())
			goto Again;
		d.val = var->u.val = 0.0;
		break;
	case 0:
		execerror("non-number read into", var->name);
		break;
	default:
		d.val = 1.0;
		break;
	}
	var->type = VAR;
	push(d);
}
